"""
Settings collector implementation.
"""

from __future__ import annotations

from typing import Tuple, Optional, Dict, Any, Iterable, Type

from .exceptions import SC_ConfigError, SC_WeirdBugError
from .manager import SC_LoadersManager
from .setting import SC_Setting
from .value import SC_Value, SC_DefaultValue


ScopesKeyType = Optional[Tuple[str, ...]]
ScopesType = Optional[Dict[ScopesKeyType, Type["SettingsCollector"]]]


class _SettingsCollectorMeta(type):
    """
    Metaclass for `SettingsCollector`.
    """

    _default_config_data = {
        # Prefix for settings' names
        "prefix": None,
        # Separator in settings' names.
        "sep": "__",
        # A configurable sequence of loaders to use or skip (see
        # `exclude`).
        "loaders": None,
        # What to do with `loaders`
        # (`False` = use them; `True` = use all others).
        "exclude": True,
        # Load data from all loaders, not just the first successful one.
        "load_all": False,
        # If set to `True`, this loads all settings when the first one is
        # requested. If `False`, they are loaded only when they are requested.
        # This does not affect the automatically reloaded settings.
        "greedy_load": True,
    }

    def __new__(metacls, name, bases, namespace, **kwargs):
        """
        Create and return a new `SettingsCollector` (sub)class.
        """
        class SC_Settings:
            """
            Settings' definitions moved from the main class.
            """

        class SC_Values:
            """
            Actual settings' values.
            """

        class SC_Data:
            """
            Configuration for the current scope.

            This is a separate class to avoid populating `SettingsCollector`
            with too many values and potentially causing a conflict with
            user-defined settings.
            """
            # The name of the current class's scope.
            scope_name: Optional[str] = None
            # Parent scope.
            parent: Optional[Type[SettingsCollector]] = None
            # Root scope
            root: Optional[Type[SettingsCollector]] = None
            # Were settings loaded greedily?
            greedy_loaded: bool = False
            # Children scopes (only valid in the root).
            scopes: ScopesType = dict()

        result = super().__new__(metacls, name, bases, namespace, **kwargs)

        # We want each subclass to have its own `SC_Values` and `SC_Data`.
        result.SC_Settings = SC_Settings  # type: ignore
        result.SC_Values = SC_Values  # type: ignore
        result.SC_Data = SC_Data  # type: ignore

        result.SC_Data.root = result

        result._process_config()
        result._check_bad_names()
        result._expand_defaults()
        result._create_sc_values()

        return result

    def __repr__(cls):
        try:
            result = cls.SC_Data.root.__name__
        except AttributeError:
            result = "<UNKNOWN Settings Collector>"
        if cls.SC_Data.scope_name:
            result += f"({repr(cls.SC_Data.scope_name)})"
        return result

    def _process_config(cls):
        """
        Populate config class with missing defaults, creating it if missing.
        """
        class SC_Config:
            """
            Configuration for the collector.

            All children scopes will share the same one.
            """

        try:
            config = cls.SC_Config
        except AttributeError:
            config = cls.SC_Config = SC_Config

        defined_values = set(dir(config)) - set(dir(object))
        for name, value in cls._default_config_data.items():
            if name not in defined_values:
                setattr(config, name, value)

    def _is_bad_name(cls, name: str) -> bool:
        """
        Return if `name` should not be used as a setting name.
        """
        return (
            # Prefix "SC_" is reserved for Settings Collector's special items.
            name.startswith("SC_")
            # Using the separator can lead to confusing results.
            or name.startswith(cls.SC_Config.sep)
        )

    def _check_bad_names(cls) -> None:
        """
        Raise `SC_ConfigError` if any of configured settings has a bad name.
        """
        bad_names = [
            name
            for name in dir(cls)
            if (
                isinstance(getattr(cls, name), (SC_Setting, SC_Value))
                and cls._is_bad_name(name)
            )
        ]
        if bad_names:
            raise SC_ConfigError(
                f"these settings need to be renamed:"
                f" {', '.join(sorted(bad_names))}",
            )

    def _expand_defaults(cls) -> None:
        """
        Expand simple defaults to normal `SC_Setting` definitions.
        """
        defaults = getattr(cls, "defaults", None)
        if isinstance(defaults, dict):
            delete_defaults = True
            for name, value in defaults.items():
                super().__setattr__(name, SC_Setting(value))
                if name == "defaults":
                    delete_defaults = False
            if delete_defaults:
                delattr(cls, "defaults")

    def _create_sc_values(cls) -> None:
        """
        Populate `SC_Values` subclass with definitions of settings.
        """
        for name in dir(cls):
            sc_setting = getattr(cls, name)
            if not isinstance(sc_setting, SC_Setting):
                continue
            sc_setting.name = name
            sc_value = SC_Value(sc_setting)
            setattr(cls.SC_Settings, name, sc_setting)  # type: ignore
            setattr(cls.SC_Values, name, sc_value)  # type: ignore
            delattr(cls, name)

    def __getattr__(cls, name: str) -> Any:
        """
        Return the `SC_Value` instance for `name`.
        """
        try:
            assert not cls._is_bad_name(name)
            sc_value = getattr(cls.SC_Values, name)  # type: ignore
            assert isinstance(sc_value, SC_Value)
        except (AssertionError, AttributeError):
            raise AttributeError(
                f"type object {repr(cls.__name__)} has no attribute"
                f" {repr(name)}",
            )
        else:
            return sc_value.getter(cls)  # type: ignore

    def __setattr__(cls, name: str, value: Any) -> None:
        """
        Set a new value for a setting.
        """
        try:
            assert not cls._is_bad_name(name)
            sc_value = getattr(cls.SC_Values, name)  # type: ignore
            assert isinstance(sc_value, SC_Value)
        except AssertionError:
            super().__setattr__(name, value)
        except AttributeError:
            raise AttributeError(f"{repr(cls)} has no setting {repr(name)}")
        else:
            sc_value.setter(cls, value)  # type: ignore


class SettingsCollector(metaclass=_SettingsCollectorMeta):
    """
    Settings collector.
    """

    defaults: Dict[str, Any] = dict()

    def __new__(  # type: ignore
        cls,
        scope_name: Optional[str] = None,
    ) -> Type[SettingsCollector]:
        """
        Return a custom scoped instance.
        """
        return cls.get_scope(scope_name)

    @classmethod
    def _get_new(
        cls,
        scope_name: Optional[str] = None,
        parent_scope: Type[SettingsCollector] = None,
    ) -> Type[SettingsCollector]:
        """
        Create and return a custom scoped instance.

        This should normally be in `__new__`, but we want that one for
        convenient use, and this one for actual creation only when needed
        (called by `settings.get_scope`).
        """
        class SettingsCollectorScope(cls):  # type: ignore
            pass

        SettingsCollectorScope._init_child_scope(scope_name, parent_scope)

        return SettingsCollectorScope

    @classmethod
    def _init_child_scope(
        cls,
        scope_name: str,
        parent_scope: Type[SettingsCollector],
    ) -> None:
        """
        Init the current `SettingsCollector` subclass as a child for a scope.
        """
        sc_values = cls.SC_Values  # type: ignore
        sc_data = cls.SC_Data  # type: ignore
        for name, sc_value in parent_scope.get_sc_values():
            setattr(sc_values, name, sc_value.clone())
        root_scope = parent_scope.SC_Data.root  # type: ignore
        sc_data.parent = parent_scope
        sc_data.scope_name = scope_name
        sc_data.root = root_scope
        sc_data.scopes = None
        sc_data.root.SC_Data.scopes[scope_name] = cls  # type: ignore
        cls.SC_Config = parent_scope.SC_Config  # type: ignore
        cls.SC_Settings = parent_scope.SC_Settings  # type: ignore
        cls.__name__ = repr(cls)

    @classmethod
    def get_sc_values(cls) -> Iterable[Tuple[str, SC_Value]]:
        """
        Return a generator of all settings' definitions.

        :return: A generator yielding `(name, sc_value)` tuples, where
            `name` is the string name of each setting and `sc_value` is an
            instance of `SC_Setting` that holds that setting's
            definition.
        """
        sc_values = cls.SC_Values  # type: ignore
        for name in dir(sc_values):
            sc_value = getattr(sc_values, name)
            if isinstance(sc_value, SC_Value):
                yield name, sc_value

    @classmethod
    def get_prefix(cls) -> str:
        """
        Return the correct prefix for settings' names.
        """
        return (
            f"{cls.SC_Config.prefix}{cls.SC_Config.sep}"
            if cls.SC_Config.prefix else
            ""
        )

    @classmethod
    def get_scope_prefix(cls) -> str:
        """
        Return the correct prefix for settings' names including the scope name.
        """
        result = cls.get_prefix()
        scope_name = cls.SC_Data.scope_name  # type: ignore
        if scope_name:
            result = f"{result}{scope_name}{cls.SC_Config.sep}"
        return result

    @classmethod
    def _get_scope(cls, scope_id: ScopesKeyType) -> Type["SettingsCollector"]:
        """
        Return the scope with the given scope ID.

        :param scope_id: A tuple of strings identifying a scope. For example,
            if `cls.SC_Config.sep = "__"` and the scope's name is `"a__bc__d"`,
            the scope ID will be `("a", "bc", "d")`.
        :return: A SettingsCollector class (not instance!) representing the
            requested scope.
        """
        if not scope_id:
            return cls

        sc_data = cls.SC_Data  # type: ignore
        if sc_data.root is None:
            raise SC_WeirdBugError(f"root in {repr(cls)}.SC_Data is None")

        scopes = sc_data.root.SC_Data.scopes  # type: ignore
        if scopes is None:
            raise SC_WeirdBugError(
                f"scopes in {repr(sc_data.root)}.SC_Data is None",
            )

        try:
            result = scopes[scope_id]
        except KeyError:
            result = cls._get_new(
                cls.SC_Config.sep.join(scope_id),
                cls._get_scope(scope_id[:-1]),
            )
            scopes[scope_id] = result
        return result

    @classmethod
    def get_scope(cls, name: Optional[str]) -> Type[SettingsCollector]:
        """
        Return a scope with the given name.
        """
        return cls._get_scope(
            tuple(name.split(cls.SC_Config.sep)) if name else None
        )

    @classmethod
    def _expand_settings_names(
        cls,
        settings_names: Optional[Iterable[str]] = None,
        as_type: Type[tuple | set] = tuple,
    ) -> tuple | set:
        """
        Return a sequence of settings' names.

        :param settings_names: A sequence of names to be returned. If set as
            `None`, all a sequence of all defined settings' names is returned.
        :param as_type: Either a `tuple` or a `set`, defining the type of the
            returned sequence.
        :return: A sequence of settings' names strings.
        """
        return as_type(
            settings_names
            if settings_names else (
                name
                for name in dir(cls.SC_Values)
                if isinstance(getattr(cls.SC_Values, name), SC_Value)
            )
        )

    @classmethod
    def _get_sc_default_values(cls) -> Dict[str, Any]:
        return {name: SC_DefaultValue for name, _ in cls.get_sc_values()}

    @classmethod
    def get_settings(
        cls,
        settings_names: Optional[Iterable[str]] = None,
        expand_names: bool = True,
    ) -> Dict[str, Any]:
        """
        Return settings' values from loaders.

        :param settings_names: Either `None` (meaning "all settings") or an
            iterable of string names of the settings to load.
        :return: A dictionary of loaded values.
        """
        result = dict()
        sc_data = cls.SC_Data  # type: ignore
        greedy_load = (
            cls.SC_Config.greedy_load
            and not sc_data.greedy_loaded
        )
        if greedy_load:
            settings_names = None
            if sc_data.parent is None:
                result.update(cls._get_sc_default_values())
        if expand_names:
            settings_names = cls._expand_settings_names(settings_names)
        result.update(SC_LoadersManager.get_settings(cls, settings_names))
        cls._assign_settings_values(result)
        if greedy_load:
            sc_data.greedy_loaded = True
        return result

    @classmethod
    def _assign_settings_values(cls, settings_values: Dict[str, Any]):
        """
        Assign values from a dictionary to `SC_Value` instances.
        """
        for name, value in settings_values.items():
            setattr(cls, name, value)

    @classmethod
    def clear_cache(cls):
        for _, sc_value in cls.get_sc_values():
            sc_value.clear_cache()
        if cls.SC_Data.scopes:
            for scope in cls.SC_Data.scopes.values():
                scope.clear_cache()



"""
Class for getting, setting, and holding the actual settings' values.
"""

from __future__ import annotations

from typing import Any, Optional, Type, TYPE_CHECKING

from .exceptions import SC_ConfigError
from .setting import SC_Setting
from .undef import SC_undef


if TYPE_CHECKING:
    from .collector import SettingsCollector  # pragma: no cover


class SC_DefaultValue:
    """
    Internal class used to represent default value without setting it.
    """


class SC_Value:
    """
    A class for holding actual values for settings.
    """

    def __init__(self, sc_setting: SC_Setting):
        self.sc_setting = sc_setting
        self.value_is_set = False
        self.value = None

    def clone(self) -> SC_Value:
        """
        Return `SC_Value` configured for the same `SC_Setting`.
        """
        return type(self)(self.sc_setting)

    def _get_default_value(self) -> Any:
        """
        Return default value if defined or raise a `ValueError` exception.
        """
        result = self.sc_setting.default
        if result is SC_undef:
            raise ValueError(
                f"setting {repr(self.sc_setting.name)} must be defined",
            )
        else:
            return result

    def cast(self, value: Optional[Any]) -> Any:
        """
        Return `value` with proper type casting.

        :param value: Value to cast as `self.value_type`.
        :raise ValueError: Raised if the casting fails and
            `self.default_on_error` is not `True`.
        :return: Value cast to the given type.
        """
        if value == SC_DefaultValue:
            return self._get_default_value()
        if self.sc_setting.value_type is None:
            return value

        has_correct_type = isinstance(value, self.sc_setting.value_type)

        if has_correct_type:
            return value
        else:
            try:
                return self.sc_setting.value_type(value)
            except Exception as ex:
                if self.sc_setting.default_on_error:
                    return self._get_default_value()
                else:
                    raise TypeError(
                        "invalid value {value} for setting {setting_name}"
                        " (it should be of type {type_name})".format(
                            value=repr(value),
                            setting_name=self.sc_setting.name,
                            type_name=self.sc_setting.value_type.__name__,
                        ),
                    ) from ex

    def getter(self, settings_collector: Type[SettingsCollector]) -> Any:
        """
        Return the value for the setting ((re)loaded if needed).
        """
        if self.sc_setting.name is None:
            # This should never happen
            raise SC_ConfigError(  # pragma: no cover
                "there be bug: setting not assigned its name",
            )

        if not self.sc_setting.no_cache and self.value_is_set:
            # Not a reloadble setting and we already had the value cached, so
            # we can just return it.
            return self.value

        # Get the value.
        parent_collector = settings_collector.SC_Data.parent  # type: ignore
        values = settings_collector.get_settings([self.sc_setting.name])

        # Return it or fall back to parent.
        try:
            self.value = self.cast(values[self.sc_setting.name])
        except KeyError:
            if parent_collector:
                sc_value = getattr(
                    parent_collector.SC_Values, self.sc_setting.name,
                )
                return sc_value.getter(parent_collector)
            else:
                return self._get_default_value()
        else:
            self.value_is_set = True
            return self.value

    def setter(
        self, settings_collector: Type[SettingsCollector], value: Any,
    ) -> None:
        """
        Set the value for the setting unless it's an auto-reloading one.
        """
        self.value = self.cast(value)
        if not self.sc_setting.no_cache:
            self.value_is_set = True

    def clear_cache(self) -> None:
        """
        Invalidate any cache that this value might hold.
        """
        self.value = None
        self.value_is_set = False



"""
Testing utilities.
"""

from contextlib import contextmanager
import os
import unittest
from unittest.mock import patch

from settings_collector import SC_EnvironLoader


# This loader was made mostly for testing, so we need to enable it.
SC_EnvironLoader.enabled = True


class TestsBase(unittest.TestCase):
    """
    The base unit tests class, used as a foundation for all other unit tests.
    """

    def setUp(self):
        """
        For future uses (common resets between runs).
        """
        pass

    def tearDown(self):
        """
        For future uses (common resets between runs).
        """
        pass


@contextmanager
def patch_env(_upper: bool = True, **kwargs):
    """
    Patch `os.environ` with `kwargs` (nothing else left in there!).

    :param _upper: If set to `True`, all the names from `kwargs` are changed to
        uppercase. This is normal for environment variables, but we do have one
        test where we want it turned off, hence having this argument.
    """
    kwargs = {
        name.upper() if _upper else name: str(value)
        for name, value in kwargs.items()
    }
    with patch.dict(os.environ, kwargs, clear=True):
        yield



"""
Loader that grabs settings from TurboGears' config.

Note: I never used TurboGears. This was made based on reading the docs and TG's
source on Github. Ping me if it's wrong, please.
"""

from typing import Any

from .base import SC_LoaderFromDict


class SC_TurboGearsLoader(SC_LoaderFromDict):
    """
    Loader that grabs settings from TurboGears's config.
    """

    @classmethod
    def get_source(cls) -> Any:
        """
        Return dictionary that with settings.
        """
        import turbogears  # type: ignore
        return turbogears.config  # pragma: no cover



"""
Loader that grabs settings from Flask's config.
"""

from typing import Any

from .base import SC_LoaderFromDict


class SC_FlaskLoader(SC_LoaderFromDict):
    """
    Loader that grabs settings from Flask's config.
    """

    name_case = str.upper

    @classmethod
    def get_source(cls) -> Any:
        """
        Return dictionary that with settings.
        """
        from flask import current_app  # type: ignore
        return current_app.config  # pragma: no cover



import io
import textwrap
import unittest
import unittest.mock

from settings_collector import SCTest, sc_test_print_expected, sc_test_run

from tests.utils import TestsBase, patch_env


class TestTest(TestsBase):

    def tearDown(self):
        SCTest.clear_cache()
        super().tearDown()

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_print(self, mock_stdout):
        sc_test_print_expected()
        self.assertEqual(
            mock_stdout.getvalue(),
            textwrap.dedent(
                """\
                SCTest__JustAString = 'covfefe'
                SCTest__AnInteger = 17
                SCTest__scope1__scope2__JustAString = 'covfefe, but scoped'
                SCTest__scope1__scope2__AnInteger = 1719
                """,
            ),
        )

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_run(self, mock_stdout):
        with patch_env(
            SCTest__JustAString="covfefe",
            SCTest__AnInteger=17,
            SCTest__scope1__scope2__JustAString="covfefe, but scoped",
            SCTest__scope1__scope2__AnInteger=1719,
        ):
            sc_test_run(False)
        self.assertEqual(
            mock_stdout.getvalue(),
            "Success: all tests have passed.\n",
        )

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_run_verbose(self, mock_stdout):
        with patch_env(
            SCTest__JustAString="covfefe",
            SCTest__AnInteger=17,
            SCTest__scope1__scope2__JustAString="covfefe, but scoped",
            SCTest__scope1__scope2__AnInteger=1719,
        ):
            sc_test_run(True)
        self.assertEqual(
            mock_stdout.getvalue(),
            textwrap.dedent(
                """\
                Test for 'JustAString' passed using SC_EnvironLoader.
                Test for 'AnInteger' passed using SC_EnvironLoader.
                Test for 'scope1__scope2__JustAString' passed using SC_EnvironLoader.
                Test for 'scope1__scope2__AnInteger' passed using SC_EnvironLoader.
                """,  # noqa: E501
            ),
        )

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_run_verbose_missing(self, mock_stdout):
        with patch_env(
            SCTest__JustAString="covfefe",
            SCTest__scope1__scope2__JustAString="covfefe, but scoped",
        ):
            sc_test_run(True)
        self.assertEqual(
            mock_stdout.getvalue(),
            textwrap.dedent(
                """\
                Test for 'JustAString' passed using SC_EnvironLoader.
                Error getting value for SCTest__AnInteger loaded from SC_EnvironLoader: setting 'AnInteger' must be defined
                Test for 'scope1__scope2__JustAString' passed using SC_EnvironLoader.
                Error getting value for SCTest__scope1__scope2__AnInteger loaded from SC_EnvironLoader: setting 'AnInteger' must be defined
                """,  # noqa: E501
            ),
        )

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_run_verbose_bad_cast(self, mock_stdout):
        self.maxDiff = None
        with patch_env(
            SCTest__JustAString="covfefe",
            SCTest__AnInteger="seventeen",
            SCTest__scope1__scope2__JustAString="covfefe, but scoped",
            SCTest__scope1__scope2__AnInteger="seventeen-nineteen",
        ):
            sc_test_run(True)
        self.assertEqual(
            mock_stdout.getvalue(),
            textwrap.dedent(
                """\
                Test for 'JustAString' passed using SC_EnvironLoader.
                Error casting value for SCTest__AnInteger loaded from SC_EnvironLoader: invalid value 'seventeen' for setting AnInteger (it should be of type int)
                Test for 'scope1__scope2__JustAString' passed using SC_EnvironLoader.
                Error casting value for SCTest__scope1__scope2__AnInteger loaded from SC_EnvironLoader: invalid value 'seventeen-nineteen' for setting AnInteger (it should be of type int)
                """,  # noqa: E501
            ),
        )

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_run_verbose_bad_value(self, mock_stdout):
        self.maxDiff = None
        with patch_env(
            SCTest__JustAString="covfefe",
            SCTest__AnInteger=19,
            SCTest__scope1__scope2__JustAString="covfefe, but scoped",
            SCTest__scope1__scope2__AnInteger=1917,
        ):
            sc_test_run(True)
        self.assertEqual(
            mock_stdout.getvalue(),
            textwrap.dedent(
                """\
                Test for 'JustAString' passed using SC_EnvironLoader.
                Invalid value received for SCTest__AnInteger loaded from SC_EnvironLoader: 19 != 17
                Test for 'scope1__scope2__JustAString' passed using SC_EnvironLoader.
                Invalid value received for SCTest__scope1__scope2__AnInteger loaded from SC_EnvironLoader: 1917 != 1719
                """,  # noqa: E501
            ),
        )



"""
Class that defines an individual setting to be collected.
"""

from __future__ import annotations

from copy import deepcopy
from typing import Optional, Dict, Type, Any

from .undef import SC_undef


class SC_Setting:
    """
    Class that defines an individual setting to be collected.

    You can compare this to Django's `django.db.models.Field`.
    """

    def __init__(
        self,
        default: Any = SC_undef,
        *,
        no_cache: bool = False,
        value_type: Optional[Type[Any]] = None,
        default_on_error: bool = True,
    ) -> None:
        """
        Initialise class instance.

        :param default: Default value for the setting.
        :param no_cache: If `True`, the value of this setting is never
            saved. Instead, it is reloaded from settings every time it is
            requested.
        :param value_type: If not `None`, this is used as a callable for the
            returned value. For example, if set to `int`, the value returned
            will be `int(value_fetched_from_settings)`.
        :param default_on_error: Fall back to `default` when casting fails due
            to invalid data.
        """
        self.default = default
        self.no_cache = no_cache
        self.value_type = value_type
        self.default_on_error = default_on_error
        self.name: Optional[str] = None

    def __copy__(self) -> SC_Setting:
        return type(self)(
            default=self.default,
            no_cache=self.no_cache,
            value_type=self.value_type,
            default_on_error=self.default_on_error,
        )

    def __deepcopy__(self, memo: Dict[int, Any]) -> SC_Setting:
        id_self = id(self)
        try:
            return memo[id_self]
        except KeyError:
            result = memo[id_self] = type(self)(
                default=deepcopy(self.default),
                no_cache=deepcopy(self.no_cache),
                value_type=deepcopy(self.value_type),
                default_on_error=deepcopy(self.default_on_error),
            )
            return result



from settings_collector import SettingsCollector, SC_Setting

from tests.utils import TestsBase, patch_env


class TestTwoSettingsCollectors(TestsBase):

    def test_two_settings_collectors(self):
        class my_settings_1(SettingsCollector):
            class SC_Config:
                prefix = "one"
            simple1 = SC_Setting()
            simple2 = SC_Setting()

        class my_settings_2(SettingsCollector):
            class SC_Config:
                prefix = "two"
            simple2 = SC_Setting()
            simple3 = SC_Setting()

        with patch_env(
            one__simple1="food", one__simple2="bard", one__simple3="nein",
            two__simple1="nope", two__simple2="foot", two__simple3="bark",
        ):
            self.assertEqual(my_settings_1.simple1, "food")
            self.assertEqual(my_settings_1.simple2, "bard")
            self.assertEqual(my_settings_2.simple2, "foot")
            self.assertEqual(my_settings_2.simple3, "bark")



from typing import Iterable, Optional, Any
import unittest.mock

from settings_collector import (
    SettingsCollector, SC_LoaderBase, SC_LoadersManager, SC_ConfigError,
)

from tests.utils import TestsBase


class _TestLoAder(SC_LoaderBase):

    @classmethod
    def get_settings(
        cls, prefix: str, settings_names: Iterable[str],
    ) -> Optional[dict[str, Any]]:
        return {"a": 11, "b": 13}


class _TestLoBder(SC_LoaderBase):

    @classmethod
    def get_settings(
        cls, prefix: str, settings_names: Iterable[str],
    ) -> Optional[dict[str, Any]]:
        return {"a": 17, "c": 19}


class TestManager(TestsBase):

    def test_get_loaders_include(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                loaders = ("loAder", "loCder")
                exclude = False

        with unittest.mock.patch(
            "settings_collector.SC_LoadersManager._loaders",
            {"loAder": _TestLoAder, "loBder": _TestLoBder},
        ):
            result = list(SC_LoadersManager._get_loaders(my_settings))
        self.assertEqual(result, [_TestLoAder])

    def test_get_loaders_exclude(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                loaders = ("loAder", "loCder")
                exclude = True

        with unittest.mock.patch(
            "settings_collector.SC_LoadersManager._loaders",
            {"loAder": _TestLoAder, "loBder": _TestLoBder},
        ):
            result = list(SC_LoadersManager._get_loaders(my_settings))
        self.assertEqual(result, [_TestLoBder])

    def test_get_loaders_include_nothing(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                loaders = tuple()
                exclude = False

        with unittest.mock.patch(
            "settings_collector.SC_LoadersManager._loaders",
            {"loAder": _TestLoAder, "loBder": _TestLoBder},
        ):
            with self.assertRaises(ValueError):
                list(SC_LoadersManager._get_loaders(my_settings))

    def test_get_loaders_include_only_unknown(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                loaders = ("loaCder", "loaDder")
                exclude = False

        with unittest.mock.patch(
            "settings_collector.SC_LoadersManager._loaders",
            {"loAder": _TestLoAder, "loBder": _TestLoBder},
        ):
            with self.assertRaises(SC_ConfigError):
                list(SC_LoadersManager._get_loaders(my_settings))

    @unittest.mock.patch("settings_collector.SC_LoadersManager._get_loaders")
    def test_get_settings_load_all(self, mock_get_loaders):
        class my_settings(SettingsCollector):
            class SC_Config:
                load_all = True

        mock_get_loaders.return_value = (
            loader for loader in (_TestLoAder, _TestLoBder)
        )
        expected = {"a": 17, "b": 13, "c": 19}

        result = SC_LoadersManager.get_settings(my_settings)

        self.assertEqual(result, expected)



"""
Loader that grabs settings from CherryPy' config.

Note: I never used CherryPy. This was made based on reading the docs. Ping me
if it's wrong, please.
"""

from copy import deepcopy
from typing import Any

from .base import SC_LoaderFromDict


class SC_CherryPyLoader(SC_LoaderFromDict):
    """
    Loader that grabs settings from CherryPy's config.
    """

    no_settings_exceptions = (ImportError, AttributeError)

    @classmethod
    def get_source(cls) -> Any:
        """
        Return dictionary that with settings.
        """
        import cherrypy  # type: ignore
        result = deepcopy(cherrypy.config)  # pragma: no cover
        result.update(cherrypy.request.app.config)  # pragma: no cover
        return result  # pragma: no cover



from settings_collector import (
    SettingsCollector, SC_Setting, SC_undef, SC_ConfigError,
)

from tests.utils import TestsBase, patch_env


class TestBasicFunctionality(TestsBase):

    def _run_test_no_default_missing(self, field_name):
        class my_settings(SettingsCollector):
            defaults = {"simple1": "foo", "simple2": SC_undef}
            simple3 = SC_Setting("bar")
            simple4 = SC_Setting()

        # Greedy loading means that fetching of _any_ field will fail if some
        # value is missing.
        with patch_env(simple1="food", simple3="bard"):
            with self.assertRaises(ValueError):
                getattr(my_settings, field_name)

    def test_no_default_missing(self):
        for field_name in ("simple1", "simple2", "simple3", "simple4"):
            self._run_test_no_default_missing(field_name)

    def test_no_default_lazy(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                greedy_load = False
            defaults = {"simple1": "foo", "simple2": SC_undef}
            simple3 = SC_Setting("bar")
            simple4 = SC_Setting()

        with patch_env(simple1="food", simple3="bard"):
            self.assertEqual(my_settings.simple1, "food")
            with self.assertRaises(ValueError):
                my_settings.simple2
            self.assertEqual(my_settings.simple3, "bard")
            with self.assertRaises(ValueError):
                my_settings.simple4

    def test_no_default_ok(self):
        class my_settings(SettingsCollector):
            defaults = {"simple1": "foo", "simple2": SC_undef}
            simple3 = SC_Setting("bar")
            simple4 = SC_Setting()

        with patch_env(simple2="food", simple4="bard"):
            self.assertEqual(my_settings.simple1, "foo")
            self.assertEqual(my_settings.simple2, "food")
            self.assertEqual(my_settings.simple3, "bar")
            self.assertEqual(my_settings.simple4, "bard")

    def test_prefix(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                prefix = "test"
                sep = "__"
            defaults = {"simple1": "foo"}
            simple2 = SC_Setting("bar")

        with patch_env(test__simple1="food", test__simple2="bard"):
            self.assertEqual(my_settings.simple1, "food")
            self.assertEqual(my_settings.simple2, "bard")

    def test_prefix_lazy(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                greedy_load = False
                prefix = "test"
                sep = "__"
            defaults = {"simple1": "foo"}
            simple2 = SC_Setting("bar")

        with patch_env(simple1="food", simple2="bard"):
            self.assertEqual(my_settings.simple1, "foo")
            self.assertEqual(my_settings.simple2, "bar")

        with patch_env(test__simple1="food", test__simple2="bard"):
            self.assertEqual(my_settings.simple1, "food")
            self.assertEqual(my_settings.simple2, "bard")

    def test_prefix_greedy(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                greedy_load = True
                prefix = "test"
                sep = "__"
            defaults = {"simple1": "foo"}
            simple2 = SC_Setting("bar")

        with patch_env(simple1="food", simple2="bard"):
            self.assertEqual(my_settings.simple1, "foo")
            self.assertEqual(my_settings.simple2, "bar")

        with patch_env(test__simple1="food", test__simple2="bard"):
            self.assertEqual(my_settings.simple1, "foo")
            self.assertEqual(my_settings.simple2, "bar")

    def test_prefix_greedy_clear_cache(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                prefix = "test"
                sep = "__"
            defaults = {"simple1": "foo"}
            simple2 = SC_Setting("bar")

        with patch_env(simple1="food", simple2="bard"):
            self.assertEqual(my_settings.simple1, "foo")
            self.assertEqual(my_settings.simple2, "bar")

        my_settings.clear_cache()

        with patch_env(test__simple1="food", test__simple2="bard"):
            self.assertEqual(my_settings.simple1, "food")
            self.assertEqual(my_settings.simple2, "bard")

    def test_cast(self):
        class my_settings(SettingsCollector):
            simple1 = SC_Setting(value_type=str)
            simple2 = SC_Setting(value_type=int)

        with patch_env(simple1="food", simple2="17"):
            self.assertEqual(my_settings.simple1, "food")
            self.assertEqual(my_settings.simple2, 17)

    def test_cached(self):
        class my_settings(SettingsCollector):
            defaults = {"simple1": "foo"}
            simple2 = SC_Setting("bar")
            int1 = SC_Setting(17, value_type=int)
            int2 = SC_Setting(19, value_type=int)
            int3 = SC_Setting(23.29, value_type=int)
            int4 = SC_Setting(31, value_type=int)

        with patch_env(simple1="food", int2=13, int4=37.41):
            self.assertEqual(my_settings.simple1, "food")
            self.assertEqual(my_settings.simple2, "bar")
            self.assertEqual(my_settings.int1, 17)
            self.assertEqual(my_settings.int2, 13)
            self.assertEqual(my_settings.int3, 23.29)
            self.assertEqual(my_settings.int4, 31)

        with patch_env(simple1="new", int2=11, int3=43):
            self.assertEqual(my_settings.simple1, "food")
            self.assertEqual(my_settings.simple2, "bar")
            self.assertEqual(my_settings.int1, 17)
            self.assertEqual(my_settings.int2, 13)
            self.assertEqual(my_settings.int3, 23.29)

        my_settings.clear_cache()

        with patch_env(int3=43):
            self.assertEqual(my_settings.int3, 43)

    def test_not_cached(self):
        class my_settings(SettingsCollector):
            nc1 = SC_Setting("default", no_cache=True)
            nc2 = SC_Setting("default", no_cache=True)
            nc3 = SC_Setting("default", no_cache=True)

        with patch_env(nc1="not default"):
            self.assertEqual(my_settings.nc1, "not default")
            self.assertEqual(my_settings.nc2, "default")
            self.assertEqual(my_settings.nc3, "default")

        with patch_env(nc1="also not default", nc2="maybe default?"):
            self.assertEqual(my_settings.nc1, "also not default")
            self.assertEqual(my_settings.nc2, "maybe default?")
            self.assertEqual(my_settings.nc3, "default")

    def test_name_case(self):
        class my_settings(SettingsCollector):
            simple1 = SC_Setting("default1")
            simple2 = SC_Setting("default2")

        with patch_env(simple1="modified1", SIMPLE2="modified2", _upper=False):
            self.assertEqual(my_settings.simple1, "default1")
            self.assertEqual(my_settings.simple2, "modified2")

    def _get_bad_name_1(self):
        class my_settings(SettingsCollector):
            SC_foo = SC_Setting()

    def _get_bad_name_2(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                sep = "xxx"

            xxxFoo = SC_Setting()

    def test_bad_names(self):
        self.assertRaises(SC_ConfigError, self._get_bad_name_1)
        self.assertRaises(SC_ConfigError, self._get_bad_name_2)

    def test_preserve_defaults_1(self):
        class my_settings(SettingsCollector):
            defaults = SC_Setting()  # not a special dict!!!

        with patch_env(defaults="ze defaults..."):
            self.assertEqual(my_settings.defaults, "ze defaults...")

    def test_preserve_defaults_2(self):
        class my_settings(SettingsCollector):
            defaults = {"defaults": "default"}

        with patch_env(defaults="ze defaults..."):
            self.assertEqual(my_settings.defaults, "ze defaults...")

    def test_setting_invalid_value(self):
        class my_settings(SettingsCollector):
            foo = SC_Setting()

        with self.assertRaises(AttributeError):
            my_settings.bar = 17



from settings_collector import SettingsCollector, SC_Setting

from tests.utils import TestsBase, patch_env


class TestScopes(TestsBase):

    def test_scope(self):
        class my_settings(SettingsCollector):
            defaults = {"simple1": "foo"}
            simple2 = SC_Setting("bar")
            int1 = SC_Setting(17, value_type=int)
            int2 = SC_Setting(19, value_type=int)

        with patch_env(
            simple1="food", x__simple1="fooled",
            y__simple2="barred",
            int1=13, x__int1=11,
            x__int2=23, y__int2=29,
        ):
            # Root
            self.assertEqual(my_settings.simple1, "food")
            self.assertEqual(my_settings.simple2, "bar")
            self.assertEqual(my_settings.int1, 13)
            self.assertEqual(my_settings.int2, 19)

            # Scope: x
            self.assertEqual(my_settings("x").simple1, "fooled")
            self.assertEqual(my_settings("x").simple2, "bar")
            self.assertEqual(my_settings("x").int1, 11)
            self.assertEqual(my_settings("x").int2, 23)

            # Scope: y
            self.assertEqual(my_settings("y").simple1, "food")
            self.assertEqual(my_settings("y").simple2, "barred")
            self.assertEqual(my_settings("y").int1, 13)
            self.assertEqual(my_settings("y").int2, 29)

    def test_deep_scope(self):
        class my_settings(SettingsCollector):
            foo = SC_Setting("foo")

        with patch_env(
            foo="empty", x__foo="x", x__y__foo="xy", x__y__z__q__foo="xyzq",
        ):
            self.assertEqual(my_settings.foo, "empty")
            self.assertEqual(my_settings("x").foo, "x")
            self.assertEqual(my_settings("x__y").foo, "xy")
            self.assertEqual(my_settings("x__y__z").foo, "xy")
            self.assertEqual(my_settings("x__y__z__q").foo, "xyzq")
            self.assertEqual(my_settings("bar").foo, "empty")

    def test_two_scopes(self):
        class my_settings(SettingsCollector):
            foo = SC_Setting("foo")

        with patch_env(x__foo="x"):
            self.assertEqual(my_settings.foo, "foo")
            self.assertEqual(my_settings("x").foo, "x")
            self.assertEqual(my_settings("bar").foo, "foo")
            self.assertEqual(my_settings("bar__x").foo, "foo")



"""
A class to register and manage all loaders.
"""

from __future__ import annotations

from typing import Type, Optional, Iterable, Dict, Any, TYPE_CHECKING

from .exceptions import SC_ConfigError, SC_NotALoader

if TYPE_CHECKING:  # pragma: no cover
    from .collector import SettingsCollector
    from .loaders.base import SC_LoaderBase


class SC_LoadersManager:
    """
    A class to register and manage all loaders.
    """

    loader_name_prefix: str = "SC_"
    loader_name_suffix: str = "Loader"
    _loaders: Dict[str, Type[SC_LoaderBase]] = dict()
    last_successful_loader: Optional[Type[SC_LoaderBase]] = None

    @classmethod
    def is_loader(cls, class_: Type[SC_LoaderBase]):
        """
        Return `True` is `class_` is a loader.
        """
        class_name = class_.__name__
        if (
            class_name.startswith(cls.loader_name_prefix)
            and class_name.endswith(cls.loader_name_suffix)
        ):
            from .loaders.base import SC_LoaderBase
            return issubclass(class_, SC_LoaderBase)
        else:
            return False

    @classmethod
    def get_loader_name(cls, class_: Type[SC_LoaderBase]):
        """
        Return loader name or raise `SC_NotALoader` exception if not a loader.
        """
        class_name = class_.__name__
        if cls.is_loader(class_):
            if cls.loader_name_prefix:
                class_name = class_name[len(cls.loader_name_prefix):]
            if cls.loader_name_suffix:
                class_name = class_name[:-len(cls.loader_name_suffix)]
            return class_name
        else:
            raise SC_NotALoader(class_name)

    @classmethod
    def register_loader(cls, loader_class: Type[SC_LoaderBase]) -> None:
        """
        Register loader defined by the given class.
        """
        try:
            loader_name = cls.get_loader_name(loader_class)
        except SC_NotALoader:
            pass
        else:
            cls._loaders[loader_name] = loader_class

    @classmethod
    def _get_loaders(
        cls,
        settings_collector: Type[SettingsCollector],
    ) -> Iterable[Type[Any]]:
        """
        Return a generator of settings loader classes.

        For arguments, see :py:meth:`get_settings`.
        """
        if (
            settings_collector.SC_Config.exclude
            or settings_collector.SC_Config.loaders
        ):
            # Find out which loaders are needed and then `yield` only those.
            include_loaders = (
                set(settings_collector.SC_Config.loaders)
                if settings_collector.SC_Config.loaders
                else set()
            )
            if settings_collector.SC_Config.exclude:
                include_loaders = set(cls._loaders) - include_loaders
            else:
                unknown_loaders = include_loaders - set(cls._loaders)
                if unknown_loaders and not include_loaders - unknown_loaders:
                    raise SC_ConfigError(
                        f"attempting to use only unknown loaders:"
                        f" {', '.join(sorted(unknown_loaders))}",
                    )
            for loader_name, loader_class in sorted(
                cls._loaders.items(), key=lambda it: it[1].priority,
            ):
                if loader_class.enabled and loader_name in include_loaders:
                    yield loader_class
        else:
            # So, you want to use only listed loaders, but then you're not
            # listing any of them? Brilliant! :-P
            raise ValueError(
                '''"You're not making any sense at all."'''
                ''' -- Captain Jack Sparrow'''
            )

    @classmethod
    def get_settings(
        cls,
        settings_collector: Type[SettingsCollector],
        settings_names: Optional[Iterable[str]] = None,
    ) -> Dict[str, Any]:
        """
        Load and return settings values as a dictionary.

        :param settings_collector: A `SettingsCollector` (sub)class for which
            the settings are being loaded.
        :param settings_names: Either `None` (meaning "all settings") or an
            iterable of string names of the settings to load.
        :raise SC_ConfigError: Raised when attempting to use only unknown
            loaders (i.e., none of the `loaders` exist and `settings.exclude`
            is set to `False`). The logic here is that you may settings.exclude
            whatever loaders you want (even those not defined, because some
            other project might have them outside of this package), but if you
            include only those loaders not available (in this package or in the
            project using it), then the whole thing becomes useless, and that's
            likely an error in the configuration of the settings collector.
        :return: A dictionary associating settings' names with their values.
        """
        result = dict()
        prefix = settings_collector.get_scope_prefix()
        for settings_loader in cls._get_loaders(settings_collector):
            settings_values = settings_loader.get_settings(
                prefix, settings_names,
            )
            if settings_values is not None:
                cls.last_successful_loader = settings_loader
                if settings_collector.SC_Config.load_all:
                    result.update(settings_values)
                else:
                    return settings_values
        return result
